from contextlib import suppress
from typing import List, Optional, Tuple, Union

import numpy as np
from matplotlib import pyplot as plt

from . import Phasor


def plot_complex(zs: Union[complex, List[complex]], color: str = None):
    if isinstance(zs, complex):
        zs = [zs]
    vectors = np.array([[z.real, z.imag] for z in zs])
    origin = np.zeros(vectors.T.shape)
    plt.quiver(*origin, vectors[:, 0], vectors[:, 1],
               color=color, angles='xy', scale_units='xy', scale=1)

    limit = max([max([abs(z.real), abs(z.imag)]) for z in zs])
    plt.xlim((-limit, limit))
    plt.ylim((-limit, limit))
    plt.grid(True, which='both')
    plt.ylabel('Im')
    plt.xlabel('Re')


def plot_phasor(*phasor_list: Phasor, color: str = None, unitary: bool = False, **kwargs):
    if unitary:
        phasor_list = [p/p.abs for p in phasor_list]
    plot_complex([p.value for p in phasor_list], color=color, **kwargs)


def plot_phasor_in_time(
    phasor: Phasor,
    time_range: Tuple[float, float, float],
    frequency: float,
    is_cos: bool = False, *,
    color: str = None,
    **kwargs
):
    t1 = time_range[0]
    t2 = time_range[1]
    num: float = 1000
    with suppress(IndexError):
        num = time_range[2] or num

    t = np.linspace(t1, t2, num)
    w = 2*np.pi*frequency
    A = phasor.abs*np.sqrt(2)
    theta = phasor.phase

    def f(t):
        if is_cos:
            return A*np.cos(w*t + theta)
        else:
            return A*np.sin(w*t + theta)

    plt.plot(t, f(t), color=color, **kwargs)
    plt.xlabel('Time (s)')
    plt.ylabel('Amplitude')
    plt.grid(True, which='both')



from typing import Tuple
from matplotlib import pyplot as plt

from . import Power


def plot_power_triangle(power: Power, *, colors: Tuple[str, str, str] = ('r', 'g', 'b'), linewidth=None, ax=None):
    if ax is None:
        ax = plt
    P, Q, S = power.active, power.reactive, power.apparent
    c1, c2, c3 = colors
    ax.plot([0, P], [0, 0], color=c1, linewidth=linewidth)
    ax.plot([P, P], [0, Q], color=c2, linewidth=linewidth)
    ax.plot([P, 0], [Q, 0], color=c3, linewidth=linewidth)
    P_str, Q_str, S_str = f'$P = {P:.2f}\;W$', f'$Q = {Q:.2f}\;VA_R$', f'$S = {S:.2f}\;VA$'
    P_str, Q_str, S_str = [s.replace('.', ',') for s in [P_str, Q_str, S_str]]

    ax.legend([P_str, Q_str, S_str])
    ax.grid()



from schemdraw import Drawing
from schemdraw import elements as elm

from .input import Input


def draw_void():
    drawing = Drawing()

    transistor = drawing.add(elm.transistors.JFetP().right().reverse())

    drawing.push()
    drawing += elm.Resistor().at(transistor.drain).up().label('Rd').length(0.93*drawing.unit)
    drawing += elm.Line().left().length(drawing.unit/2)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.SourceV().up().reverse().label('Vdd')
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/2)
    drawing += elm.Resistor().down().label('R1')
    drawing += elm.Line().down().length(0.24*drawing.unit)
    drawing += elm.Dot()

    drawing.push()
    drawing += elm.Line().to(transistor.gate)

    drawing.pop()
    drawing.push()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Capacitor().left().reverse().label('C1')
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Dot().label('Vi')

    drawing.pop()
    drawing += elm.Resistor().label('R2').length(0.93*drawing.unit)
    drawing += elm.Ground()

    drawing += elm.Resistor().label('Rs').down().at(transistor.source).length(0.78*drawing.unit)
    drawing += elm.Ground()

    drawing += elm.Dot().at(transistor.source)
    drawing += elm.Line().right().length(0.75*drawing.unit)
    drawing += elm.Capacitor().label('Cs').down().length(0.78*drawing.unit)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Dot().at(transistor.drain)
    drawing += elm.Capacitor().right().label('C2')
    drawing += elm.Dot().label('Vo')

    return drawing


def draw(amplifier_input: Input):
    Vdd = str(amplifier_input.Vdd)
    R1 = str(amplifier_input.R1)
    R2 = str(amplifier_input.R2)
    Rd = str(amplifier_input.Rd)
    Rs = str(amplifier_input.Rs)

    drawing = Drawing()

    transistor = drawing.add(elm.transistors.JFetP().right().reverse())

    drawing.push()
    drawing += elm.Resistor().at(transistor.drain).up().label(Rd).length(0.93*drawing.unit)
    drawing += elm.Line().left().length(drawing.unit/2)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.SourceV().up().reverse().label(Vdd)
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/2)
    drawing += elm.Resistor().down().label(R1)
    drawing += elm.Line().down().length(0.24*drawing.unit)
    drawing += elm.Dot()

    drawing.push()
    drawing += elm.Line().to(transistor.gate)

    drawing.pop()
    drawing.push()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Capacitor().left().reverse().label('C1')
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Dot().label('Vi')

    drawing.pop()
    drawing += elm.Resistor().label(R2).length(0.93*drawing.unit)
    drawing += elm.Ground()

    drawing += elm.Resistor().label(Rs).down().at(transistor.source).length(0.78*drawing.unit)
    drawing += elm.Ground()

    drawing += elm.Dot().at(transistor.source)
    drawing += elm.Line().right().length(0.75*drawing.unit)
    drawing += elm.Capacitor().label('Cs').down().length(0.78*drawing.unit)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Dot().at(transistor.drain)
    drawing += elm.Capacitor().right().label('C2')
    drawing += elm.Dot().label('Vo')

    return drawing



from pyelectric.electronic.amplifier import Amplifier as _Amplifier

from . import drawing
from .input import Input
from .output import Output


class Amplifier(_Amplifier):
    input: Input = Input(0, 0, 0, 0, 0, 0, 0)
    output: Output = Output(0, 0, 0, 0)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.input = Input(*args, **kwargs)

    def __call__(self) -> Output:
        Vdd = self.input.Vdd
        Rg = self.input.Rg
        Rs = self.input.Rs
        Vp = self.input.Vp
        Idss = self.input.Idss
        rd = self.input.rd
        Idq = self.input.Idq

        Vgs = ((Idq/Idss)**(1/2) - 1)*-Vp
        gm = (2*Idss/abs(Vp))*(1 - (Vgs/Vp))

        Zi = Rg
        Zo = 1/(1/Rs + 1/rd + gm)
        Req = gm/(1/rd + 1/Rs)
        Avnl = Req/(1 + Req)
        return Output(gm, Zi, Zo, Avnl)

    @staticmethod
    def draw_void():
        return drawing.draw_void()

    def draw(self):
        return drawing.draw(self.input)



from __future__ import annotations

import math
from typing import List

from .position import Position


class Conductor:
    radius: float
    position: Position

    def __init__(self, radius: float, position: Position):
        self.radius = radius
        self.position = position

    @property
    def gmr(self) -> float:
        return self.radius*math.exp(-1/4)

    def distance(self, position: Position) -> float:
        x1, y1 = self.position.x, self.position.y
        x2, y2 = position.x, position.y
        return math.sqrt((x1 - x2)**2 + (y1 - y2)**2)


def parse_conductors(conductors) -> List[Conductor]:
    conductors_list: List[Conductor] = []
    for conductor in conductors:
        if isinstance(conductor, Conductor):
            conductors_list.append(conductor)
        elif isinstance(conductor, dict):
            radius = conductor['radius']
            x, y = conductor['position']
            position = Position(x, y)
            conductors_list.append(Conductor(radius, position))
    return conductors_list



from termcolor import colored

from .bar import Bar


def complex_to_str(c: complex) -> str:
    return f'{c.real:.4e} + {c.imag:.4e}j'


class Line:
    admittance: complex
    bar1: Bar
    bar2: Bar
    amperage: complex = 0 + 0j
    power: complex = 0 + 0j
    power_reverse: complex = 0 + 0j

    def __init__(self, bar1: Bar, bar2: Bar, *, admittance: complex = None, impedance: complex = None):
        assert admittance is not None or impedance is not None
        self.bar1 = bar1
        self.bar2 = bar2
        if admittance is not None:
            self.admittance = admittance
        elif impedance is not None:
            self.admittance = 1/impedance

    def __str__(self) -> str:
        S12 = colored(complex_to_str(self.power), 'green')
        S21 = colored(complex_to_str(self.power_reverse), 'green')
        S = colored(complex_to_str(self.power + self.power_reverse), 'green')
        bar1 = colored(self.bar1.name, 'blue')
        bar2 = colored(self.bar2.name, 'blue')
        return f'{bar1} ━━ {bar2}: S = ({S}), S{bar1}{bar2} = ({S12}), S{bar2}{bar1} = ({S21})'



from abc import ABC, abstractmethod, abstractstaticmethod

from schemdraw import Drawing

from . import drawing
from .input import Input
from .output import Output


class Amplifier(ABC):
    input: Input
    output: Output

    @abstractstaticmethod
    def draw_void() -> Drawing:
        pass

    @abstractmethod
    def draw(self) -> Drawing:
        pass

    def draw_equivalent(self) -> Drawing:
        output = self()
        return drawing.draw_equivalent_model(output)

    def draw_equivalent_void(self) -> Drawing:
        return drawing.draw_equivalent_model_void()

    @abstractmethod
    def __call__(self) -> Output:
        pass



from schemdraw import Drawing
from schemdraw import elements as elm

from .input import Input


def draw_void():
    drawing = Drawing()

    transistor = drawing.add(elm.transistors.JFetP().reverse().right())
    drawing += elm.Line().left()

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.Line().up().length(drawing.unit*0.3)
    drawing += elm.Resistor().up().label('Rg')
    drawing += elm.SourceV().up().label('Vgg').length(drawing.unit/2)
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Capacitor().left()
    drawing += elm.Dot().label('Vi')

    drawing += elm.Resistor().up().at(transistor.drain).label('Rd')
    drawing += elm.SourceV().label('Vdd').reverse().length(drawing.unit/2)
    drawing += elm.Ground().left()

    drawing += elm.Line().at(transistor.source).down().length(drawing.unit/8)
    drawing += elm.Ground()

    drawing += elm.Dot().at(transistor.drain)
    drawing += elm.Capacitor().right().label('C2')
    drawing += elm.Dot().label('Vo')

    return drawing


def draw(amplifier_input: Input):
    Vdd = amplifier_input.Vdd
    Vgg = amplifier_input.Vgg
    Rd = amplifier_input.Rd
    Rg = amplifier_input.Rg

    drawing = Drawing()

    transistor = drawing.add(elm.transistors.JFetP().reverse().right())
    drawing += elm.Line().left()

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.Line().up().length(drawing.unit*0.3)
    drawing += elm.Resistor().up().label(str(Rg))
    drawing += elm.SourceV().up().label(str(Vgg)).length(drawing.unit/2)
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Capacitor().left()
    drawing += elm.Dot().label('Vi')

    drawing += elm.Resistor().up().at(transistor.drain).label(str(Rd))
    drawing += elm.SourceV().label(str(Vdd)).reverse().length(drawing.unit/2)
    drawing += elm.Ground().left()

    drawing += elm.Line().at(transistor.source).down().length(drawing.unit/8)
    drawing += elm.Ground()

    drawing += elm.Dot().at(transistor.drain)
    drawing += elm.Capacitor().right().label('C2')
    drawing += elm.Dot().label('Vo')

    return drawing


from typing import List

from pyelectric.electronic.amplifier.input import Input as _Input


class Input(_Input):
    Vcc: float
    Rb: float
    Re: float
    beta: float
    ro: float

    def __init__(self, Vcc: float, Rb: float, Re: float, beta: float, ro: float):
        self.Vcc = Vcc
        self.Rb = Rb
        self.Re = Re
        self.beta = beta
        self.ro = ro

    @staticmethod
    def get_parameter_names() -> List[str]:
        return ['Vcc', 'beta', 'ro', 'Rb', 'Re']



import math
from typing import List

import numpy as np

from .bar import Bar
from .bar.load_bar import LoadBar
from .bar.regulator_bar import RegulatorBar
from .bar.slack_bar import SlackBar
from .line import Line


class Circuit:
    bars: List[Bar]
    lines: List[Line]
    power_base: float
    __y_bus_array: np.ndarray
    __voltage_array: np.ndarray
    __power_esp_array: np.ndarray

    def __init__(self, bars: List[Bar], lines: List[Line], power_base: float = 1):
        self.bars = bars
        self.lines = lines
        self.power_base = power_base
        self.bar_powers_to_pu()
        self.__y_bus_array = self.get_y_bus_array()
        self.__voltage_array = self.get_voltage_array()
        self.__power_esp_array = self.get_power_esp_array()

    def __str__(self) -> str:
        bars = '\n'.join([str(bar) for bar in self.bars])
        lines = '\n'.join([str(line) for line in self.lines])
        return f'{bars}\n{lines}'

    def bar_powers_to_pu(self):
        for bar in self.bars:
            bar.power /= self.power_base
            if isinstance(bar, RegulatorBar):
                bar.active_power /= self.power_base
            elif isinstance(bar, LoadBar):
                bar.active_power /= self.power_base
                bar.reactive_power /= self.power_base

    def get_bar_index(self, bar: Bar) -> int:
        for i, b in enumerate(self.bars):
            if b == bar:
                return i
        raise ValueError(f'Bar {bar} not found')

    def get_y_bus_array(self) -> np.ndarray:
        y_bus = np.zeros((len(self.bars), len(self.bars)), dtype=complex)
        for line in self.lines:
            bar1_index = self.get_bar_index(line.bar1)
            bar2_index = self.get_bar_index(line.bar2)
            y_bus[bar1_index, bar2_index] = -line.admittance
            y_bus[bar2_index, bar1_index] = -line.admittance

            y_bus[bar1_index, bar1_index] = sum(
                [l.admittance for l in self.lines if l.bar1 == line.bar1 or l.bar2 == line.bar1])

            y_bus[bar2_index, bar2_index] = sum(
                [l.admittance for l in self.lines if l.bar1 == line.bar2 or l.bar2 == line.bar2])

        return y_bus

    def get_power_esp_array(self) -> np.ndarray:
        power_g = np.zeros((len(self.bars)), dtype=complex)
        power_d = np.zeros((len(self.bars)), dtype=complex)
        for i, bar in enumerate(self.bars):
            if isinstance(bar, LoadBar):
                power_d[i] = bar.power
            elif isinstance(bar, RegulatorBar):
                power_g[i] = bar.active_power
        power_esp = power_g - power_d
        return power_esp

    def get_voltage_array(self) -> np.ndarray:
        voltages = np.ones(len(self.bars), dtype=complex)
        for i, bar in enumerate(self.bars):
            if isinstance(bar, SlackBar):
                voltages[i] = bar.voltage
            elif isinstance(bar, RegulatorBar):
                voltages[i] = bar.voltage_module
        return voltages

    def get_bar_power(self, bar: Bar) -> complex:
        voltages = self.__voltage_array
        y_bus = self.__y_bus_array
        i = self.get_bar_index(bar)
        summation = sum([y_bus[i, j]*voltages[j]
                        for j in range(len(voltages)) if i != j])
        S = (voltages[i].conjugate()*(y_bus[i, i]
             * voltages[i] + summation)).conjugate()
        return S

    def get_bar_voltage(self, bar: Bar) -> complex:
        voltages = self.__voltage_array
        y_bus = self.__y_bus_array
        power_esp = self.__power_esp_array
        i = self.get_bar_index(bar)
        I = power_esp[i].conjugate()/voltages[i].conjugate()
        summation = sum([y_bus[i, j]*voltages[j]
                        for j in range(len(voltages)) if i != j])
        V = (I - summation)/y_bus[i, i]
        return V

    def solve(self, repeat: int = 1, max_error: float = None):
        if max_error is None:
            for _ in range(repeat):
                self.__update_bar_voltages()
        else:
            while True:
                voltages_old = self.__voltage_array.copy()
                self.__update_bar_voltages()
                voltages_new = self.__voltage_array
                error = voltages_new - voltages_old
                real_error, imag_error = np.abs(error.real), np.abs(error.imag)
                if (np.all(real_error < max_error) and np.all(imag_error < max_error)):
                    break

        self.__update_bar_powers()
        self.__update_line_amperages()
        self.__update_line_powers()

    def __update_bar_voltages(self):
        voltages = self.__voltage_array
        power_esp = self.__power_esp_array

        for i, bar in enumerate(self.bars):
            if isinstance(bar, LoadBar):
                voltages[i] = self.get_bar_voltage(bar)
            elif isinstance(bar, RegulatorBar):
                Q = self.get_bar_power(bar).imag
                P = power_esp[i].real
                power_esp[i] = P + 1j*Q
                V = self.get_bar_voltage(bar)
                V_real = math.sqrt(bar.voltage_module**2 - V.imag**2)
                voltages[i] = V_real + V.imag*1j

        self.__voltage_array = voltages
        for i, bar in enumerate(self.bars):
            bar.voltage = voltages[i]

    def __update_bar_powers(self):
        for i, bar in enumerate(self.bars):
            if isinstance(bar, SlackBar):
                bar.power = self.get_bar_power(bar)
            elif isinstance(bar, RegulatorBar):
                bar.power = self.__power_esp_array[i]

    def __update_line_amperages(self):
        y_bus = self.__y_bus_array
        y = y_bus*(np.identity(len(y_bus))*2 - 1)
        for line in self.lines:
            bar1 = line.bar1
            bar2 = line.bar2
            bar1_index = self.get_bar_index(bar1)
            bar2_index = self.get_bar_index(bar2)
            I = (bar1.voltage - bar2.voltage)*y[bar1_index, bar2_index]
            line.amperage = I

    def __update_line_powers(self):
        for line in self.lines:
            I = line.amperage

            V1 = line.bar1.voltage
            S12 = V1*I.conjugate()
            line.power = S12

            V2 = line.bar2.voltage
            S21 = V2*(-I).conjugate()
            line.power_reverse = S21



import math

from pyelectric.ac.power import Power


def reactance(voltage: float, power: Power, new_factor: float) -> float:
    new_reactive_power = math.tan(math.acos(new_factor)) * power.active
    Q = power.reactive - new_reactive_power
    return voltage**2 / Q


def capacitance(reactance: float, frequency: float) -> float:
    return 1/(2*math.pi*frequency*reactance)



from typing import List

from pyelectric.electronic.amplifier.output import Output as _Output


class Output(_Output):
    gm: float
    Zi: float
    Zo: float
    Avnl: float

    def __init__(self, gm: float, Zi: float, Zo: float, Avnl: float) -> None:
        self.gm = gm
        self.Zi = Zi
        self.Zo = Zo
        self.Avnl = Avnl

    def __str__(self) -> str:
        return '\n'.join([
            f"gm = {self.gm:.2f} S",
            f"Zi = {self.Zi:.2f} Ω",
            f"Zo = {self.Zo:.2f} Ω",
            f"Avnl = {self.Avnl:.2f}"
        ])

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def get_parameter_names() -> List[str]:
        return ['gm', 'Zi', 'Zo', 'Avnl']



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyelectric",
    version="0.3.2",
    author="luanws",
    author_email="luan.w.silveira@gmail.com",
    description="Python package for electric circuits",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/luanws/pyelectric",
    packages=setuptools.find_packages(include=['pyelectric', 'pyelectric.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy',
        'matplotlib',
    ],
    python_requires='>=3.7',
)



from pyelectric.electronic.amplifier import Amplifier as _Amplifier

from . import drawing
from .input import Input
from .output import Output


class Amplifier(_Amplifier):
    input: Input = Input(0, 0, 0, 0, 0)
    output: Output = Output(0, 0, 0, 0)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.input = Input(*args, **kwargs)

    def __call__(self) -> Output:
        Vcc = self.input.Vcc
        Rb = self.input.Rb
        Re = self.input.Re
        beta = self.input.beta
        ro = self.input.ro

        Ibq = (Vcc - 0.7)/(Rb + Re*(beta + 1))
        Ieq = (beta + 1)*Ibq

        re = 26e-3/Ieq
        Zb = beta*(Re + re)
        Zi = 1/((1/Rb) + (1/Zb))
        Zo = 1/((1/(beta*re)) + (1/Re) + (1/ro) + (1/re))
        Avnl = Re/(Re + re)

        return Output(re, Zi, Zo, Avnl)

    @staticmethod
    def draw_void():
        return drawing.draw_void()

    def draw(self):
        return drawing.draw(self.input)
